package org.jboss.cache.buddyreplication;

import org.jboss.cache.CacheSPI;
import org.jboss.cache.Fqn;
import org.jboss.cache.Cache;
import org.jboss.cache.notifications.annotation.CacheListener;
import org.jboss.cache.notifications.annotation.CacheBlocked;
import org.jboss.cache.notifications.annotation.CacheUnblocked;
import org.jboss.cache.notifications.event.Event;
import org.jboss.cache.commands.remote.DataGravitationCleanupCommand;
import org.jboss.cache.util.CachePrinter;
import org.jboss.cache.util.TestingUtil;
import org.jboss.cache.util.internals.replicationlisteners.ReplicationListener;
import static org.testng.AssertJUnit.*;
import org.testng.annotations.BeforeTest;
import org.testng.annotations.Test;
import org.testng.annotations.BeforeClass;

import javax.transaction.TransactionManager;
import java.util.List;

/**
 * @author Mircea.Markus@jboss.com
 */
@Test(groups = "functional", testName = "buddyreplication.Buddy3NodesNoPoolWithDataGravitationTest")
public class Buddy3NodesNoPoolWithDataGravitationTest extends AbstractNodeBasedBuddyTest
{
   private Fqn fqn = Fqn.fromString("test");

   private String key = "key";
   private String value = "value";


   @BeforeClass
   public void createCaches() throws Exception
   {
      caches = createCaches(3, false, true);
   }

   public void testDataGravitationDontKillOwner() throws Exception
   {
      Fqn fqn = Fqn.fromString("/test");
      Fqn backupFqn = fqnTransformer.getBackupFqn(caches.get(0).getLocalAddress(), fqn);

      TestingUtil.dumpCacheContents(caches);

      caches.get(0).put(fqn, key, value);

      TestingUtil.dumpCacheContents(caches);

      assertEquals("Value should exist", value, caches.get(0).get(fqn, key));

      TestingUtil.dumpCacheContents(caches);

      // use exists instead of get() to prevent going up the interceptor stack
      assertTrue("Should be false", !caches.get(1).exists(fqn));
      assertTrue("Should be false", !caches.get(2).exists(fqn));

      assertFalse("Should be false", caches.get(0).exists(backupFqn));
      assertTrue("Value be true", caches.get(1).exists(backupFqn));
      assertFalse("Should be false", caches.get(2).exists(backupFqn));

      ReplicationListener replListener = ReplicationListener.getReplicationListener(caches.get(1));
      replListener.expect(DataGravitationCleanupCommand.class);

      // according to data gravitation, a call to *any* cache should retrieve the data, and move the data to the new cache.
      assertEquals("Value should have gravitated", value, caches.get(2).get(fqn, key));
      replListener.waitForReplicationToOccur();

      TestingUtil.dumpCacheContents(caches);

      // now lets test the eviction part of gravitation
      Fqn newBackupFqn = fqnTransformer.getBackupFqn(caches.get(2).getLocalAddress(), fqn);

      // use exists instead of get() to prevent going up the interceptor stack
      assertTrue("Should be false", !caches.get(0).exists(fqn));
      assertTrue("Should be false", !caches.get(1).exists(fqn));

      // the old backup should no longer exist
      assertFalse("Should be null", caches.get(0).exists(backupFqn));
      assertFalse("Should be null", caches.get(1).exists(backupFqn));
      assertFalse("Should be null", caches.get(2).exists(backupFqn));

      // and the backup should now exist in caches.get(2)'s buddy which is caches.get(0)
      assertEquals("Value should exist", value, caches.get(0).get(newBackupFqn, key));
      assertFalse("Should be null", caches.get(1).exists(newBackupFqn));
      assertFalse("Should be null", caches.get(2).exists(newBackupFqn));
   }



   public void testTransactionsCommit() throws Exception
   {


      caches.get(0).put(fqn, key, value);
      Fqn oldBackupFqn = Fqn.fromString("/" + BuddyManager.BUDDY_BACKUP_SUBTREE + "/" + fqnTransformer.getGroupNameFromAddress(caches.get(0).getLocalAddress()) + "/test");
      Fqn newBackupFqn = Fqn.fromString("/" + BuddyManager.BUDDY_BACKUP_SUBTREE + "/" + fqnTransformer.getGroupNameFromAddress(caches.get(2).getLocalAddress()) + "/test");

//      dumpCacheContents(caches);

      TransactionManager txman = caches.get(2).getTransactionManager();


      assertTrue(caches.get(0).exists(fqn));
      assertTrue(!caches.get(1).exists(fqn));
      assertTrue(!caches.get(2).exists(fqn));
      assertTrue(!caches.get(0).exists(oldBackupFqn));
      assertTrue(caches.get(1).exists(oldBackupFqn));
      assertTrue(!caches.get(2).exists(oldBackupFqn));
      assertTrue(!caches.get(0).exists(newBackupFqn));
      assertTrue(!caches.get(1).exists(newBackupFqn));
      assertTrue(!caches.get(2).exists(newBackupFqn));


      txman.begin();

//      dumpCacheContents(caches);
      // will cause gravitation
      caches.get(2).get(fqn, key);
//      dumpCacheContents(caches);

      assertTrue(caches.get(0).exists(fqn));
      assertTrue(!caches.get(1).exists(fqn));
      assertTrue(caches.get(2).exists(fqn));
      assertTrue(!caches.get(0).exists(oldBackupFqn));
      assertTrue(caches.get(1).exists(oldBackupFqn));
      assertTrue(!caches.get(2).exists(oldBackupFqn));
      assertTrue(!caches.get(0).exists(newBackupFqn));
      assertTrue(!caches.get(1).exists(newBackupFqn));
      assertTrue(!caches.get(2).exists(newBackupFqn));

      txman.commit();
      cleanupDelay(); // cleanup commands are async

//      dumpCacheContents(caches);

      assertTrue(!caches.get(0).exists(fqn));
      assertTrue(!caches.get(1).exists(fqn));
      assertTrue(caches.get(2).exists(fqn));
      assertTrue(!caches.get(0).exists(oldBackupFqn));
      assertTrue(!caches.get(1).exists(oldBackupFqn));
      assertTrue(!caches.get(2).exists(oldBackupFqn));
      assertTrue(caches.get(0).exists(newBackupFqn));
      assertTrue(!caches.get(1).exists(newBackupFqn));
      assertTrue(!caches.get(2).exists(newBackupFqn));

//      dumpCacheContents(caches);

      assertNoLocks(caches);
   }

   private void cleanupDelay()
   {
      TestingUtil.sleepThread(250);
   }

   public void testTransactionsRollback() throws Exception
   {
      TestingUtil.dumpCacheContents(caches.get(0));
      TestingUtil.dumpCacheContents(caches.get(1));
      TestingUtil.dumpCacheContents(caches.get(2));

      caches.get(0).put(fqn, key, value);
      Fqn oldBackupFqn = Fqn.fromString("/" + BuddyManager.BUDDY_BACKUP_SUBTREE + "/" + fqnTransformer.getGroupNameFromAddress(caches.get(0).getLocalAddress()) + "/test");
      Fqn newBackupFqn = Fqn.fromString("/" + BuddyManager.BUDDY_BACKUP_SUBTREE + "/" + fqnTransformer.getGroupNameFromAddress(caches.get(2).getLocalAddress()) + "/test");

//      dumpCacheContents(caches);

      TransactionManager txman = caches.get(2).getTransactionManager();


      assertTrue(caches.get(0).exists(fqn));
      assertTrue(!caches.get(1).exists(fqn));
      assertTrue(!caches.get(2).exists(fqn));
      assertTrue(!caches.get(0).exists(oldBackupFqn));
      assertTrue(caches.get(1).exists(oldBackupFqn));
      assertTrue(!caches.get(2).exists(oldBackupFqn));
      assertTrue(!caches.get(0).exists(newBackupFqn));
      assertTrue(!caches.get(1).exists(newBackupFqn));
      assertTrue(!caches.get(2).exists(newBackupFqn));


      txman.begin();

//      dumpCacheContents(caches);
      caches.get(2).get(fqn, key);
//      dumpCacheContents(caches);

      assertTrue(caches.get(0).exists(fqn));
      assertTrue(!caches.get(1).exists(fqn));
      assertTrue(caches.get(2).exists(fqn));
      assertTrue(!caches.get(0).exists(oldBackupFqn));
      assertTrue(caches.get(1).exists(oldBackupFqn));
      assertTrue(!caches.get(2).exists(oldBackupFqn));
      assertTrue(!caches.get(0).exists(newBackupFqn));
      assertTrue(!caches.get(1).exists(newBackupFqn));
      assertTrue(!caches.get(2).exists(newBackupFqn));

      txman.rollback();

//      dumpCacheContents(caches);

      assertTrue(caches.get(0).exists(fqn));
      assertTrue(!caches.get(1).exists(fqn));
      assertTrue(!caches.get(2).exists(fqn));
      assertTrue(!caches.get(0).exists(oldBackupFqn));
      assertTrue(caches.get(1).exists(oldBackupFqn));
      assertTrue(!caches.get(2).exists(oldBackupFqn));
      assertTrue(!caches.get(0).exists(newBackupFqn));
      assertTrue(!caches.get(1).exists(newBackupFqn));
      assertTrue(!caches.get(2).exists(newBackupFqn));

      assertNoLocks(caches);
   }

   public void testSubtreeRetrieval() throws Exception
   {
      Fqn fqn = Fqn.fromString("/test");
      Fqn fqn2 = Fqn.fromString("/test/subtree");

      Fqn backupFqn = fqnTransformer.getBackupFqn(caches.get(0).getLocalAddress(), fqn);
      Fqn backupFqn2 = fqnTransformer.getBackupFqn(caches.get(0).getLocalAddress(), fqn2);

      caches.get(0).put(fqn, key, value);
      caches.get(0).put(fqn2, key, value);

      // test backup replication to buddy
      assertEquals(value, caches.get(0).get(fqn, key));
      assertEquals(value, caches.get(0).get(fqn2, key));
      assertEquals(value, caches.get(1).get(backupFqn, key));
      assertEquals(value, caches.get(1).get(backupFqn2, key));

      assertTrue(!caches.get(0).exists(backupFqn));
      assertTrue(!caches.get(0).exists(backupFqn2));
      assertTrue(!caches.get(1).exists(fqn));
      assertTrue(!caches.get(1).exists(fqn2));
      assertTrue(!caches.get(2).exists(fqn));
      assertTrue(!caches.get(2).exists(fqn2));
      assertTrue(!caches.get(2).exists(backupFqn));
      assertTrue(!caches.get(2).exists(backupFqn2));

      assertNoLocks(caches);

      // gravitate to 2:
      ReplicationListener replListener1 = ReplicationListener.getReplicationListener(caches.get(0));
      replListener1.expect(DataGravitationCleanupCommand.class);
      caches.get(2).getNode(fqn);  // expectWithTx entire subtree to gravitate.
      replListener1.waitForReplicationToOccur(); // cleanup commands are async

      Fqn newBackupFqn = fqnTransformer.getBackupFqn(caches.get(2).getLocalAddress(), fqn);
      Fqn newBackupFqn2 = fqnTransformer.getBackupFqn(caches.get(2).getLocalAddress(), fqn2);

      assertEquals(value, caches.get(2).get(fqn, key));
      assertTrue(caches.get(2).exists(fqn2));
      assertEquals(value, caches.get(0).get(newBackupFqn, key));
      assertTrue(caches.get(0).exists(newBackupFqn2));

      assertTrue(!caches.get(2).exists(newBackupFqn));
      assertTrue(!caches.get(2).exists(newBackupFqn2));
      assertTrue(!caches.get(0).exists(fqn));
      assertTrue(!caches.get(0).exists(fqn2));
      assertTrue(!caches.get(1).exists(fqn));
      assertTrue(!caches.get(1).exists(fqn2));
      assertTrue(!caches.get(1).exists(newBackupFqn));
      assertTrue(!caches.get(1).exists(newBackupFqn2));

      TestingUtil.dumpCacheContents(caches);

      for (CacheSPI<Object, Object> cache : caches)
      {
         assertTrue(!cache.exists(backupFqn));
         assertTrue(!cache.exists(backupFqn2));
      }

      assertNoLocks(caches);
   }

   private void testDataOwner(boolean optimistic) throws Exception
   {
      List<CacheSPI<Object, Object>> caches = createCaches(1, 2, false, true, optimistic);
      cachesTL.set(caches);

      // add some stuff on the primary
      CacheSPI dataOwner = caches.get(0);
      CacheSPI buddy = caches.get(1);

      dataOwner.put(fqn, key, value);

      assert dataOwner.peek(fqn, false) != null : "Should have data";
      assert dataOwner.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(buddy.getLocalAddress())), false) != null : "Should have backup node for buddy";
      assert dataOwner.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(dataOwner.getLocalAddress())), false) == null : "Should NOT have backup node for self!";

      assert buddy.peek(fqn, false) == null : "Should not have data";
      assert buddy.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(buddy.getLocalAddress())), false) == null : "Should NOT have backup node for self!";
      assert buddy.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(dataOwner.getLocalAddress())), false) != null : "Should have backup node for buddy";
      assert buddy.peek(fqnTransformer.getBackupFqn(dataOwner.getLocalAddress(), fqn), false) != null : "Should have backup data";

      // now do a gravitate call.
      assert buddy.get(fqn, key).equals(value) : "Data should have gravitated!";
      // gravitation cleanup calls are async.
      cleanupDelay();

      assert buddy.peek(fqn, false) != null : "Should have data";
      assert buddy.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(dataOwner.getLocalAddress())), false) != null : "Should have backup node for buddy";
      assert buddy.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(buddy.getLocalAddress())), false) == null : "Should NOT have backup node for self!";

      assert dataOwner.peek(fqn, false) == null : "Should not have data";
      assert dataOwner.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(dataOwner.getLocalAddress())), false) == null : "Should NOT have backup node for self!";
      assert dataOwner.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(buddy.getLocalAddress())), false) != null : "Should have backup node for buddy";
      assert dataOwner.peek(fqnTransformer.getBackupFqn(buddy.getLocalAddress(), fqn), false) != null : "Should have backup data";
   }

   public void testStaleRegionOnDataOwner() throws Exception
   {
      // add some stuff on the primary
      CacheSPI first = caches.get(0);
      CacheSPI second = caches.get(1);
      CacheSPI third = caches.get(2);

      first.put(fqn, key, value);

      assert first.peek(fqn, false) != null : "Should have data";
      assert first.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(third.getLocalAddress())), false) != null : "Should have backup node for second";
      assert first.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(first.getLocalAddress())), false) == null : "Should NOT have backup node for self!";

      assert second.peek(fqn, false) == null : "Should not have data";
      assert second.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(second.getLocalAddress())), false) == null : "Should NOT have backup node for self!";
      assert second.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(first.getLocalAddress())), false) != null : "Should have backup node for second";
      assert second.peek(fqnTransformer.getBackupFqn(first.getLocalAddress(), fqn), false) != null : "Should have backup data";

      // now do a gravitate call.
      assert second.get(fqn, key).equals(value) : "Data should have gravitated!";
      // gravitation cleanup calls are async.
      cleanupDelay();

      assert second.peek(fqn, false) != null : "Should have data";
      assert second.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(first.getLocalAddress())), false) != null : "Should have backup node for second";
      assert second.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(second.getLocalAddress())), false) == null : "Should NOT have backup node for self!";

      assert third.peek(fqn, false) == null : "Should not have data";
      assert third.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(third.getLocalAddress())), false) == null : "Should NOT have backup node for self!";
      assert third.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(second.getLocalAddress())), false) != null : "Should have backup node for second";
      assert third.peek(fqnTransformer.getBackupFqn(second.getLocalAddress(), fqn), false) != null : "Should have backup data";
   }

   public void testStaleRegionOnBuddy() throws Exception
   {
      Fqn fqn = Fqn.fromString("/a/b/c");
      Object key = "key", value = "value";

      // add some stuff on the primary
      CacheSPI dataOwner = caches.get(0);
      CacheSPI buddy = caches.get(1);
      CacheSPI thirdInstance = caches.get(2);

      assertIsBuddy(dataOwner, buddy, true);
      assertIsBuddy(buddy, thirdInstance, true);
      assertIsBuddy(thirdInstance, dataOwner, true);

      dataOwner.put(fqn, key, value);

      assert dataOwner.peek(fqn, false) != null : "Should have data";
      assert dataOwner.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(thirdInstance.getLocalAddress())), false) != null : "Should have backup node for buddy";
      assert dataOwner.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(dataOwner.getLocalAddress())), false) == null : "Should NOT have backup node for self!";
      assert dataOwner.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(buddy.getLocalAddress())), false) == null : "Should NOT have backup node for 2nd instance!";

      assert buddy.peek(fqn, false) == null : "Should not have data";
      assert buddy.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(buddy.getLocalAddress())), false) == null : "Should NOT have backup node for self!";
      assert buddy.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(dataOwner.getLocalAddress())), false) != null : "Should have backup node for buddy";
      assert buddy.peek(fqnTransformer.getBackupFqn(dataOwner.getLocalAddress(), fqn), false) != null : "Should have backup data";

      // now do a gravitate call.
      assert thirdInstance.get(fqn, key).equals(value) : "Data should have gravitated!";

      // gravitation cleanup calls are async.
      TestingUtil.sleepThread(250);

      assert thirdInstance.peek(fqn, false) != null : "Should have data";
      assert thirdInstance.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(buddy.getLocalAddress())), false) != null : "Should have backup node for buddy";
      assert thirdInstance.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(thirdInstance.getLocalAddress())), false) == null : "Should NOT have backup node for self!";

      assert dataOwner.peek(fqn, false) == null : "Should not have data";
      assert dataOwner.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(dataOwner.getLocalAddress())), false) == null : "Should NOT have backup node for self!";
      assert dataOwner.peek(Fqn.fromRelativeElements(BuddyManager.BUDDY_BACKUP_SUBTREE_FQN, fqnTransformer.getGroupNameFromAddress(thirdInstance.getLocalAddress())), false) != null : "Should have backup node for buddy";
      assert dataOwner.peek(fqnTransformer.getBackupFqn(thirdInstance.getLocalAddress(), fqn), false) != null : "Should have backup data";
      assert buddy.peek(fqn, false) == null : "Should not have data";
      assert buddy.peek(fqn.getParent(), false) == null : "Should not have any part of the data";
      assert buddy.peek(fqnTransformer.getBackupFqn(dataOwner.getLocalAddress(), fqn), false) == null : "Should NOT have backup data";
   }


   public void testConcurrency() throws Exception
   {
      Fqn f = Fqn.fromString("/a/b/c");
      String k = "k";
      String v = "v";

      for (int i = 0; i < 10; i++)
      {
         caches.get(0).put(Fqn.fromRelativeElements(f, i), k, v);
         assert v.equals(caches.get(1).get(Fqn.fromRelativeElements(f, i), k));
         assert v.equals(caches.get(1).get(Fqn.fromRelativeElements(f, i), k));
      }
   }


   @CacheListener
   public static class CacheBlockListener
   {
      private int blocks = 0;

      @CacheBlocked
      public void processBlock(Event e)
      {
         if (e.isPre())
         {
            synchronized (this)
            {
               blocks++;
               notifyAll();
            }
         }
      }

      @CacheUnblocked
      public void processUnblock(Event e)
      {
         if (e.isPre())
         {
            synchronized (this)
            {
               blocks--;
               notifyAll();
            }
         }
      }

      public void blockUntilAllCachesAreUnblocked(long maxWait) throws InterruptedException
      {
         synchronized (this)
         {
            if (blocks > 1)
            {
               wait(maxWait);
            }
            // try returning anyway?
            blocks = 0;
            /*
            if (blocks > 1)
               throw new RuntimeException("Timed out waiting for unblocks.  Number of blocks = " + blocks);
            if (blocks == 1) blocks = 0;
            */
         }
      }
   }
   
   @Test (enabled = true, dependsOnMethods = {"testConcurrency","testStaleRegionOnBuddy","testStaleRegionOnDataOwner", "testDataGravitationDontKillOwner", "testTransactionsCommit", "testTransactionsRollback", "testSubtreeRetrieval"})
   public void testCompleteStateSurvival() throws Exception
   {
      CacheBlockListener blockListener = new CacheBlockListener();

      caches.get(0).addCacheListener(blockListener);
      caches.get(1).addCacheListener(blockListener);
      caches.get(2).addCacheListener(blockListener);

      caches.get(0).put("/0", "key", "value");
      caches.get(1).put("/1", "key", "value");
      caches.get(2).put("/2", "key", "value");

//      TestingUtil.sleepThread(getSleepTimeout());
      //TestingUtil.sleepThread(caches.get(0).getConfiguration().getStateRetrievalTimeout() * 3);
      blockListener.blockUntilAllCachesAreUnblocked(caches.get(0).getConfiguration().getStateRetrievalTimeout() * 3);

      caches.get(2).stop();

      blockListener.blockUntilAllCachesAreUnblocked(caches.get(0).getConfiguration().getStateRetrievalTimeout() * 5);

      assertEquals("value", caches.get(0).get("/2", "key"));

      blockListener.blockUntilAllCachesAreUnblocked(caches.get(0).getConfiguration().getStateRetrievalTimeout() * 5);

      caches.get(1).stop();

      // cache[0] is all thats left!!

      assertEquals("value", caches.get(0).get("/0", "key"));

      try
      {
         assertEquals("value", caches.get(0).get("/1", "key"));
      }
      catch (RuntimeException e)
      {
         // may barf the first time around since we are unable to contact our buddy and store this data.
         assertEquals(IllegalArgumentException.class, e.getCause().getClass());
      }

      // now try the assertion again since the local gravitation would have worked.

      assertEquals("value", caches.get(0).get("/1", "key"));
      assertEquals("value", caches.get(0).get("/2", "key"));
      caches.get(0).stop();

      caches = createCaches(3, false, true);

      waitForSingleBuddy(caches);
   }

}
