/*
 * JBoss, Home of Professional Open Source
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.buddyreplication;

import org.jboss.cache.CacheSPI;
import org.jboss.cache.Fqn;
import org.jboss.cache.commands.remote.DataGravitationCleanupCommand;
import org.jboss.cache.util.TestingUtil;
import org.jboss.cache.util.internals.replicationlisteners.ReplicationListener;
import org.jgroups.JChannel;
import org.jgroups.protocols.DISCARD;
import static org.testng.AssertJUnit.*;
import static org.testng.AssertJUnit.assertEquals;
import org.testng.annotations.Test;

import java.util.List;

/**
 * Tests behaviour when data owners fail - essentially this tests data gravitation
 *
 * @author <a href="mailto:manik AT jboss DOT org">Manik Surtani (manik AT jboss DOT org)</a>
 */
@Test(groups = "functional", testName = "buddyreplication.Buddy3NodesWithFailoverTest")
public class Buddy3NodesWithFailoverTest extends BuddyReplicationTestsBase
{
   protected boolean optimisticLocks = false;
   protected String key = "key";
   protected String value = "value";
   BuddyFqnTransformer fqnTransformer = new BuddyFqnTransformer();

   public void testSubtreeRetrieval() throws Exception
   {
      List<CacheSPI<Object, Object>> caches = createCaches(3, false, true, optimisticLocks);
      cachesTL.set(caches);

      Fqn fqn = Fqn.fromString("/test");
      Fqn fqn2 = Fqn.fromString("/test/subtree");

      Fqn backupFqn = fqnTransformer.getBackupFqn(caches.get(0).getLocalAddress(), fqn);
      Fqn backupFqn2 = fqnTransformer.getBackupFqn(caches.get(0).getLocalAddress(), fqn2);

      caches.get(0).put(fqn, key, value);
      caches.get(0).put(fqn2, key, value);

      // test backup replication to buddy
      assertEquals(value, caches.get(0).get(fqn, key));
      assertEquals(value, caches.get(0).get(fqn2, key));
      assertEquals(value, caches.get(1).get(backupFqn, key));
      assertEquals(value, caches.get(1).get(backupFqn2, key));

      assertTrue(!caches.get(0).exists(backupFqn));
      assertTrue(!caches.get(0).exists(backupFqn2));
      assertTrue(!caches.get(1).exists(fqn));
      assertTrue(!caches.get(1).exists(fqn2));
      assertTrue(!caches.get(2).exists(fqn));
      assertTrue(!caches.get(2).exists(fqn2));
      assertTrue(!caches.get(2).exists(backupFqn));
      assertTrue(!caches.get(2).exists(backupFqn2));

      assertNoLocks(caches);

      // gravitate to 2:
      ReplicationListener replListener1 = ReplicationListener.getReplicationListener(caches.get(0));
      replListener1.expect(DataGravitationCleanupCommand.class);
      caches.get(2).getNode(fqn);  // expectWithTx entire subtree to gravitate.
      replListener1.waitForReplicationToOccur(); // cleanup commands are async

      Fqn newBackupFqn = fqnTransformer.getBackupFqn(caches.get(2).getLocalAddress(), fqn);
      Fqn newBackupFqn2 = fqnTransformer.getBackupFqn(caches.get(2).getLocalAddress(), fqn2);

      assertEquals(value, caches.get(2).get(fqn, key));
      assertTrue(caches.get(2).exists(fqn2));
      assertEquals(value, caches.get(0).get(newBackupFqn, key));
      assertTrue(caches.get(0).exists(newBackupFqn2));

      assertTrue(!caches.get(2).exists(newBackupFqn));
      assertTrue(!caches.get(2).exists(newBackupFqn2));
      assertTrue(!caches.get(0).exists(fqn));
      assertTrue(!caches.get(0).exists(fqn2));
      assertTrue(!caches.get(1).exists(fqn));
      assertTrue(!caches.get(1).exists(fqn2));
      assertTrue(!caches.get(1).exists(newBackupFqn));
      assertTrue(!caches.get(1).exists(newBackupFqn2));

      TestingUtil.dumpCacheContents(caches);

      for (CacheSPI<Object, Object> cache : caches)
      {
         assertTrue(!cache.exists(backupFqn));
         assertTrue(!cache.exists(backupFqn2));
      }

      assertNoLocks(caches);
   }

   @Test
   public void testDataGravitationKillOwner() throws Exception
   {
      List<CacheSPI<Object, Object>> cachePool = createCaches(3, false, true);
      Fqn fqn = Fqn.fromString("/test");
      Fqn backupFqn = fqnTransformer.getBackupFqn(cachePool.get(0).getLocalAddress(), fqn);

      TestingUtil.dumpCacheContents(cachePool);

      cachePool.get(0).put(fqn, key, value);

      TestingUtil.dumpCacheContents(cachePool);

      assertEquals("Value should exist", value, cachePool.get(0).get(fqn, key));

      TestingUtil.dumpCacheContents(cachePool);

      // use exists instead of get() to prevent going up the interceptor stack
      assertTrue("Should be false", !cachePool.get(1).exists(fqn));
      assertTrue("Should be false", !cachePool.get(2).exists(fqn));

      assertFalse("Should be false", cachePool.get(0).exists(backupFqn));
      assertTrue("Value be true", cachePool.get(1).exists(backupFqn));
      assertFalse("Should be false", cachePool.get(2).exists(backupFqn));

      // forcefully kill data owner.
      amendDiscard(cachePool.get(0), cachePool.get(1), 2, true);
      waitForSingleBuddy(cachePool.get(1), cachePool.get(2));

      TestingUtil.dumpCacheContents(cachePool);
      // assert that the remaining cachePool have picked new buddies.  Cache 1 should have cache 2's backup data.
      assert cachePool.get(1).peek(fqnTransformer.getBackupRoot(cachePool.get(2).getLocalAddress()), false) != null : "Should have new buddy's backup root.";
      assert cachePool.get(1).peek(fqnTransformer.getBackupRoot(cachePool.get(1).getLocalAddress()), false) == null : "Should not have self as a backup root.";
      assert cachePool.get(1).peek(fqnTransformer.getBackupRoot(cachePool.get(0).getLocalAddress()), false) == null : "Should not have dead node as a backup root.";
      assert cachePool.get(1).peek(Fqn.fromRelativeElements(fqnTransformer.getDeadBackupRoot(cachePool.get(0).getLocalAddress()), 1), false) != null : "Should have dead node as a defunct backup root.";

      assert cachePool.get(2).peek(fqnTransformer.getBackupRoot(cachePool.get(2).getLocalAddress()), false) == null : "Should not have self as a backup root.";
      assert cachePool.get(2).peek(fqnTransformer.getBackupRoot(cachePool.get(1).getLocalAddress()), false) != null : "Should have new buddy's backup root.";
      assert cachePool.get(2).peek(fqnTransformer.getBackupRoot(cachePool.get(0).getLocalAddress()), false) == null : "Should not have dead node as a backup root.";
      assert cachePool.get(2).peek(Fqn.fromRelativeElements(fqnTransformer.getDeadBackupRoot(cachePool.get(0).getLocalAddress()), 1), false) == null : "Should not have dead node as a defunct backup root.";

      ReplicationListener replListener = ReplicationListener.getReplicationListener(cachePool.get(1));
      replListener.expect(DataGravitationCleanupCommand.class);

      // according to data gravitation, a call to *any* cache should retrieve the data, and move the data to the new cache.
      assertEquals("Value should have gravitated", value, cachePool.get(2).get(fqn, key));
      replListener.waitForReplicationToOccur();

      TestingUtil.dumpCacheContents(cachePool);

      // now lets test the eviction part of gravitation
      Fqn newBackupFqn = fqnTransformer.getBackupFqn(cachePool.get(2).getLocalAddress(), fqn);

      // use exists instead of get() to prevent going up the interceptor stack
      assert cachePool.get(1).peek(fqnTransformer.getBackupRoot(cachePool.get(2).getLocalAddress()), false) != null : "Should have new buddy's backup root.";
      assert cachePool.get(1).peek(fqnTransformer.getBackupRoot(cachePool.get(1).getLocalAddress()), false) == null : "Should not have self as a backup root.";
      assert cachePool.get(1).peek(fqnTransformer.getBackupRoot(cachePool.get(0).getLocalAddress()), false) == null : "Should not have dead node as a backup root.";
      assert cachePool.get(1).peek(Fqn.fromRelativeElements(fqnTransformer.getDeadBackupRoot(cachePool.get(0).getLocalAddress()), 1), false) == null : "Should not have dead node as a defunct backup root.";
      assert cachePool.get(1).peek(fqnTransformer.getDeadBackupRoot(cachePool.get(0).getLocalAddress()), false) == null : "Should not have dead node as a defunct backup root.";

      assert cachePool.get(2).peek(fqnTransformer.getBackupRoot(cachePool.get(2).getLocalAddress()), false) == null : "Should not have self as a backup root.";
      assert cachePool.get(2).peek(fqnTransformer.getBackupRoot(cachePool.get(1).getLocalAddress()), false) != null : "Should have new buddy's backup root.";
      assert cachePool.get(2).peek(fqnTransformer.getBackupRoot(cachePool.get(0).getLocalAddress()), false) == null : "Should not have dead node as a backup root.";
      assert cachePool.get(2).peek(Fqn.fromRelativeElements(fqnTransformer.getDeadBackupRoot(cachePool.get(0).getLocalAddress()), 1), false) == null : "Should not have dead node as a defunct backup root.";
      assertTrue("Should be false", !cachePool.get(1).exists(fqn));

      // the old backup should no longer exist
      assertFalse("Should be null", cachePool.get(1).exists(backupFqn));
      assertFalse("Should be null", cachePool.get(2).exists(backupFqn));
   }

   private void amendDiscard(CacheSPI cacheToKill, CacheSPI anotherCache, int finalExpectedClusterSize, boolean enable)
   {
      JChannel channel = (JChannel) cacheToKill.getRPCManager().getChannel();
      DISCARD discard = (DISCARD) channel.getProtocolStack().findProtocol(DISCARD.class);
      if (discard != null)
      {
         discard.setDiscardAll(enable);
         TestingUtil.blockUntilViewReceived(anotherCache, finalExpectedClusterSize, 10000, false);
         // an extra second, for some reason we need this.
         TestingUtil.sleepThread(1000);
      }
   }
}
