/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.loader;

import org.jboss.cache.Fqn;
import org.jboss.cache.CacheSPI;
import org.jboss.cache.factories.UnitTestCacheConfigurationFactory;
import org.jboss.cache.config.CacheLoaderConfig;
import org.jboss.cache.transaction.DummyTransactionManager;
import org.jboss.cache.util.TestDbPropertiesFactory;
import static org.testng.AssertJUnit.*;
import org.testng.annotations.Test;
import org.testng.annotations.AfterTest;
import org.hsqldb.jdbc.jdbcDataSource;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NameNotFoundException;
import java.util.Properties;

/**
 * This test runs cache loader tests using Database as the cache loader store.
 * The default test is configured using MySQL.
 * The server and database configuration is read from a properties file located at
 * /etc/cache-jdbc.properties.
 * <p/>
 * The appropriate JDBC driver (i.e mysql-connector-java-3.0.10-stable-bin.jar)
 * must be in the lib directory for this test to run successfuly
 *
 * @author <a href="hmesha@novell.com">Hany Mesha</a>
 * @version <tt>$Revision: 7332 $</tt>
 */
@Test(groups = {"functional"}, testName = "loader.JDBCCacheLoaderHsqldbDSTest")
public class JDBCCacheLoaderHsqldbDSTest extends CacheLoaderTestsBase
{
   //private String old_factory = null;
   private final String FACTORY = "org.jboss.cache.transaction.DummyContextFactory";
   private final String JNDI_NAME = "java:/DerbyDS";

   protected void configureCache(CacheSPI cache) throws Exception
   {
      if (ds == null)
      {
         System.setProperty(Context.INITIAL_CONTEXT_FACTORY, FACTORY);
         DummyTransactionManager.getInstance();

         Context context = new InitialContext();
         try
         {
            Object obj = context.lookup(JNDI_NAME);
            assertNull(JNDI_NAME + " not bound", obj);
         }
         catch (NameNotFoundException n)
         {
            // expected
         }

         prop = TestDbPropertiesFactory.getTestDbProperties();

         ds = new jdbcDataSource();
         ds.setDatabase(prop.getProperty("cache.jdbc.url"));
         ds.setUser("sa");

         context.bind(JNDI_NAME, ds);
         assertNotNull(JNDI_NAME + " bound", context.lookup(JNDI_NAME));
      }
      String props = "cache.jdbc.datasource =" + JNDI_NAME + "\n" +
            "cache.jdbc.node.type=" + prop.getProperty("cache.jdbc.node.type") + "\n" +
            "cache.jdbc.sql-concat= 1 || 2" + "\n" +
            "cache.jdbc.table.name=jbosscache" + "\n" +
            "cache.jdbc.table.primarykey=jbosscache_pk";
      CacheLoaderConfig config = UnitTestCacheConfigurationFactory.buildSingleCacheLoaderConfig(false, "", "org.jboss.cache.loader.JDBCCacheLoader",
            props, false, true, false, false, false);
      cache.getConfiguration().setCacheLoaderConfig(config);
      cache.create();
   }

   private Properties prop;

   private jdbcDataSource ds;

   @AfterTest
   protected void distroyDbAfterTest() throws Exception
   {
      Properties icProps = new Properties();
      icProps.setProperty(Context.INITIAL_CONTEXT_FACTORY, FACTORY);
      Context ctx = new InitialContext(icProps);
      ctx.unbind(JNDI_NAME);
      TestDbPropertiesFactory.shutdownInMemoryDatabase(prop);
   }

   public void testLargeObject()
   {
      try
      {
         String key = "LargeObj";
         // create an object with size bigger than 4k (k=1024 bytes)
         StringBuilder text = new StringBuilder("LargeObject");
         while (text.toString().getBytes().length < (1024 * 100))
         {
            text.append(text);
         }
         String initialValue = text.toString();
         // insert it into the cache loader
         loader.remove(Fqn.fromString("/"));

         Object retVal = loader.put(FQN, key, initialValue);
         assertNull(retVal);
         addDelay();
         // load the object from the cache loader and validate it
         assertEquals(initialValue, (String) loader.get(FQN).get(key));
         // update the object and validate it
         String updatedValue = initialValue.concat(("UpdatedValue"));
         retVal = loader.put(FQN, key, updatedValue);
         assertEquals(initialValue, (String) retVal);
         assertEquals(updatedValue, (String) loader.get(FQN).get(key));
      }
      catch (Exception e)
      {
         fail(e.toString());
      }
   }

   public void testTransactionRollback() throws Exception
   {
      // no-op
   }

   public void testIntegratedTransactionRollback() throws Exception
   {
      // no-op
   }

   public void testRemoveData()
   {
      super.testRemoveData();
   }
}
